#ifndef ftp_header_included
#define ftp_header_included


//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.


///@file
///  Defines and interface to the FTP utilities.

#include <cxxtls/file.h>

namespace cxxtls
{

class FTP
    ///  This class defines an interface to the ftp program.
    ///  Basically it provides filesystem like queries and
    ///  operations that operate on a remote machine using
    ///  the ftp program to do the work.
    ///
{
public:

    typedef std::string string;

    enum constants
    {
       max_ftp_line_length=2048    ///<  max length of directory lines
    };


    static string error();
	///  Get the text of the most recent error -- if any.


    static bool parse_ftp_dir_text(char const*  buffer,
				   FileStatus*  statbuf,
				   FileName*    filename
				  );
	///<
	///  Convert a string provided by ftp for directory entries into
	///  file statuses.  Return true if it is a valid entry, false if
	///  the text is not correctly formatted.  Also, this function returns
	///  a pointer to the filename part of the ftp directory text --
	///  so you don't have to figure out where it is yourself.
	///
	///  @return true if the buffer contains a syntactly correct FTP directory output log line
	///  @param buffer   line of text provided by the ftp program (given the dir command) for a file status display
	///  @param statbuf  a FileStatus object into which the buffer is parsed
	///  @param filename the name of the file on the ftp dir output line
	///
	
	
    struct FileInfo
	///  A FileInfo is a combination of the name of a file and its
	///  FileStatus.
    {
	FileName    name_;
	FileStatus  status_;
	
	FileInfo(FileName const &name, FileStatus const& status)
	: name_(name),
	  status_(status)
	{
	  ///  construct a fileinfo object given its name and status
	}
	
	FileInfo()
	{
	}
	
	
    };


    static int stat_matching(char const *host,
			     char const *user,
			     char const *password,
			     char const *pattern,
			     std::list<FileInfo> *output);
      ///<
      /// Find a list of FileName's matching a specified file name pattern
      /// on a remote host using ftp -- get descriptors of those files
      /// which contain the file name and status.  This is much more
      /// efficient than using find_matching followed by a bunch of file_stat
      /// calls.
      ///
      ///  @return error code
      ///  @param host the name of the hostmachine to which a query is being done
      ///  @param username the name of the user id to pass to the ftp program
      ///  @param password the password required to get the ftp program to work
      ///  @param pattern the file name pattern to search for
      ///  @param output the list of fileInfo data structures requested by this call
      ///

    static int stat_matching(string          host,
			     string          user,
			     string          password,
			     string          pattern,
			     std::list<FileInfo>* l)
      ///
      /// Find a list of FileName's matching a specified file name pattern
      /// on a remote host using ftp -- get descriptors of those files
      /// which contain the file name and status.  This is much more
      /// efficient than using find_matching followed by a bunch of file_stat
      /// calls.
      ///
      ///  @return error code
      ///  @param host the name of the hostmachine to which a query is being done
      ///  @param user the name of the user id to pass to the ftp program
      ///  @param password the password required to get the ftp program to work
      ///  @param pattern the file name pattern to search for
      ///  @param l the list of fileInfo data structures requested by this call
      ///
    {
      return stat_matching(host.c_str(),
			   user.c_str(),
			   password.c_str(),
			   pattern.c_str(),
			   l
			  );
    }


    static bool get(string   host,
		    string   user,
		    string   password,
		    FileName pathname,
		    string   to_file=""
		   );
	///< Use ftp binary transfer to retrieve a file from a
	///  remote system.  You can optionally specify a file
	///  name into which to store the result.  By default the
	///  file node name will be the name of the retrieved file.
	///  @return error code
	///  @param host the name of the host machine for ftp activity
	///  @param user the user name on the host machine
	///  @param password the needed password
	///  @param pathname the file to replace
	///  @param to_file  local file to store the retrieved file.

    static bool put(string   host,
		    string   user,
		    string   password,
		    FileName pathname,
		    string   from_file=""
		   );
	///< Use ftp binary transfer to store a file to a
	///  remote system.  You can optionally specify a file
	///  name from which the file comes.  By default the
	///  file node name of the target file will be assumed to be
	///  the name of the source file (in the current directory).
        ///  @return error code
        ///  @param host the name of the host machine for ftp activity
        ///  @param user the user name on the host machine
        ///  @param password the needed password
        ///  @param pathname the file to get
	///  @param from_file  the file to put to the ftp server

    static bool del(string   host,
		    string   user,
		    string   password,
		    FileName pathname
		   );

	///< Use ftp to delete a file on a remote host.
        ///  @return error code
        ///  @param host the name of the host machine for ftp activity
        ///  @param user the user name on the host machine
        ///  @param password the needed password
	///  @param pathname the file to delete

};
} // namespace cxxtls
#endif
