#include <cxxtls/foreach.h>
#include <cxxtls/simple_regex.h>
#include <cxxtls/charSetTranslator.h>
#include <list>
#include <memory>
#include <map>
#include <portable_io.h>

#ifdef __GNUC__
#   if __GNUC__ > 4 || ( __GNUC__ == 4 &&  __GNUC_MINOR__ > 2 )
#       define USE_muSED
#   endif
#else
#   define USE_muSED
#endif


///@file cxxtls/muSED.h
/// Define namespace @ref muSED which is a "micro sed", where sed is the
/// Unix "String EDitor" program.  

namespace cxxtls
{

namespace muSED
    /// The namespace that contains the micro-SED interpreter definitions.
    ///   
    /// See the following classes and
    /// functions of interest:
    ///
    ///   @li muSED::apply() -- applies muSED scripts to input sequences.  Various flavors.
    ///   @li muSED::CompiledScript -- Representation of a compiled script and parsers for
    ///                                statements.
    ///   @li muSED::ScriptRuntimeContext -- Representation of the instantateous state of the execution
    ///                                      system as a script is applied to an input stream.
    ///   @li muSED::Statement -- a representation of the a statement in the muSED language.
    ///
{
    using namespace cxxtls;

    // micro sed engine.  Sed is the "string editor" program on unix.
    // muSED lets you execute simplified sed scripts on string containers (vectors, 
    // arrays, lists, dequeues, etc.



    struct ScriptRuntimeContext;  // compiled script plus state variables and buffers

//========================================================================================
     
     
     template<class StringIterator1, class StringIterator2>
     void skipBlanks(StringIterator1 &firstChar,
                     StringIterator2 &lastChar
                    )
     {
        /// move firstChar past any blanks, tabs, etc
        /// but not past lastChar
        ///
        /// @param[in,out] firstChar  Pointer to the first character in a range
        /// @param[in]     lastChar   Pointer past the end of the range
    
    
        while(    (firstChar != lastChar)
              &&  (*firstChar <= ' ')
             )
             ++firstChar;
     }

//========================================================================================

     void expandEscapes(std::string &input)
         ///  translate standard escape sequences, such as \\n, into the single character
         ///  form (such as 0x0a).
         ;

//========================================================================================
     template<class StringIterator1, class StringIterator2>
     bool parseEscapedString(char            delim, 
                             std::string    &output,
                             StringIterator1 &firstChar,
                             StringIterator2 &lastChar
                            )
     {

        ///  Parse for a string until the delimiter is reached, handling delims.
        ///  All other \ sequences will be left unmodified, but will be skipped
        ///  over properly.
        ///
        ///  NOTE:  the firstChar should not be the leading delimiter of a 
        ///         delimited group.  For example, if you are dealing with a
        ///         regex, first char should be past the leading delimiter,
        ///         as in the following:
        ///@code
        ///
        ///    /regex/
        ///     ^
        ///     fc
        ///
        ///@endcode

        char c;

        StringIterator1 savedFirst = firstChar;

        while(    (firstChar != lastChar)
              &&  (c = *firstChar)
              &&  (c != delim)
             )
        {
           if(c == '\\')
           {
              ++firstChar;

              if(firstChar == lastChar)
              {
                 firstChar = savedFirst;  
                 return false;  //  failure
              }

              output += '\\';
              output += *firstChar++;

           }
           else
           {
              output += c;

              ++firstChar;
           }
        }


        return true;

     }  // parseEscapedString()

//========================================================================================



    struct Filter

        /// A (possibly empty) criteria for deciding to execute a line in a sed script.
        /// The criteria include
        ///
        ///    @li does the current line number match the script requirement
        ///    @li is the current line the last line of the file
        ///    @li does the current line match a regular expression
        ///
        /// The filter can be any of the above, or none.
        ///
        ///@note   Filter logic inversion is handled by FilterableStatement,
        ///        note here

    {
        // line number, regex, lastLine or other filter

        enum type { none, regex, lineNumber, lastLine };

        type type_;

        SimpleRegex regex_;

        size_t line_;


        Filter()
        : type_(none)
        , line_(0)
        {
        }

        virtual ~Filter()
        {
        }

        Filter &operator= (Filter const &rhs)
        {
           type_ = rhs.type_;

           line_ = rhs.line_;

           if(type_ == regex)
           {
              // slight optimization, don't ccopy the regex unnecessarily

              regex_ = rhs.regex_;
           }

           return *this;

        }

        bool passes(size_t lineNo, bool isLast, std::string const &line)
            ///<  Determine if the current line of text passes the
            ///   filter.  If the filter type is none, all of them
            ///   do.
            ///
            ///   @param[in] lineNo  The number of the current line of text
            ///   @param[in] isLast  A boolean indicating that this line
            ///                      is the last line of the input stream
            ///   @param[in] line    The text of the current line
            ///
            ///   @return true to indicate that the line matches the filter
            ///                
            ;

        template<class StringIterator>
        bool parse(StringIterator &first, StringIterator const &last)
           /// parse a string into *this.
           ///
           /// @param[in,out] first  Pointer to the next input character;
           /// @param[in]     last   Pointer past the end of valid input
           ///
           /// @return true to indicate the successfull parsing of a filter.
           ///  
           /// Note that returning false only means that there was no filter,
           /// not that an error occurred -- although it might mean that!
           ///
           /// Leading blanks will be skipped.  If an actual error occurrs,
           /// The first non-blank character will be the location where this
           /// function leaves 'first', which is an in/out parameter.
        {
            skipBlanks(first, last);

            if(first == last)
               return false;

            type_ = none;

            StringIterator savedFirst = first;

            char c = *first;

            switch(c)
            {
               case '/':
                   {
                      //
                      //  parse a regex, return false if not found (also restore cp
                      //

                      std::string buf;

                      ++first;

                      if(!parseEscapedString('/', buf, first, last))
                      {
                         first = savedFirst;
                         return false;
                      }

                      type_ = regex;
                      regex_ = buf;
                      ++first;  // skip the trailing /

                   }
                   return true; // if we get here!

               case '$':
                   type_ = lastLine;
                   ++first;
                   return true;

               default:
                   if(!isdigit(c))
                      return false;   // does not indicate an error
                   {
                      
                      char buffer[256];

                      char *p = &buffer[0];
                      char *e = &buffer[sizeof(buffer)-1];

                      while(   (p != e) 
                            && (first != last)
                            && (c = *first)
                            && isdigit(c)
                           )
                      {
                          *p++ = c;
                          ++first;
                      }

                      *p++ = 0;

                      int lineNo = -1;

                      sscanf(buffer, "%d", &lineNo);
                      
                      if(lineNo == -1)
                      {
                         first = savedFirst;
                         return false;
                      }

                      line_ = lineNo;
                      type_ = lineNumber;

                   }
                   return true; // if we get here
            }

            return true;
        }

        virtual std::string debugString() const;




    };  // struct Filter
//========================================================================================
    struct Statement;

    enum LineRangeActivationState
       ///  Statements with two conditionals as prefixes define a range of lines, and
       ///  this enumeration type defines the states of activation that such a statement
       ///  can be in.
    {
        neverExecuted=0,   ///<  The state before the first active line is found 
        inRange,           ///<  On or after first active line, but not on the last line
        lastLine,          ///<  On the last line
    };


    typedef std::map<Statement*, LineRangeActivationState> ActivationMap;

//========================================================================================

    struct Statement
        /// A statement as defined by the sed language -- except for branches
        /// and labels which are not supported.
        ///
        /// A sed script is a collection of sed language statements which will 
        /// all be applied to all input lines from a stream -- unless those
        /// statements are suppressed by a line selection filter.  In muSED,
        /// branches and labels are not supported, nor is the test command which
        /// is a variant of the branch command.
        ///
        /// The goal of a sed script is to transform an input stream of lines to
        /// an output stream of lines.  In the program "sed", the default logic
        /// is to copy stdin to stdout without modifcation, but in muSED, the
        /// default logic is to delete the stream and produce no output.  All
        /// desired output must be produced using the 'p' command to print the
        /// current pattern buffer (ie the most recently read input line) to
        /// it's output location.  Thus, the following is the do nothing script
        /// in muSED:
        ///
        ///@code
        ///
        ///   p
        ///
        ///@endcode
        ///
        ///
        /// sed Statements have an optional prefix and statement body.
        ///
        /// The prefix contains 0, 1, or 2 filters used to decide if the 
        /// statement should be executed on the current input line.  
        /// 
        /// 0 selection
        /// filters mean that the statement body applies to all lines to
        /// be processed.  
        ///
        /// 1 filter prefix means that the statement body
        /// applies ONLY to lines that match the filter.  
        ///
        /// 2 filter
        /// prefixes define a range of lines.  The body is applied to all
        /// lines from the first line that matches the first filter to the
        /// first line that matches the second filter.  The command will
        /// be applied to all ines in the range.  You can use a nested
        /// command with its own filter to avoid having the command
        /// applied to the last line.
        ///
        /// filters look basically like this:
        ///
        ///@code
        ///    <filterExpression>[!]
        ///@endcode
        ///
        /// That is, any single filter can have it's logic inverted by
        /// simply appending!.  For example, the filter that selects
        /// only line 12, looks like this:
        ///@code
        ///
        ///   12
        ///
        ///@endcode
        /// To select all lines BUT 12, do this:
        ///@code
        ///
        ///   12!
        ///
        ///@endcode
        ///
        /// statement bodies include primary command characters and parameters
        ///
        ///@code
        ///
        ///    s    -- string substitute
        ///    p    -- copy the current pattern buffer to the output sequence (i.e. "print" it)
        ///    P    -- print the leading fragment of the current input line, up to the first \\n
        ///            (which could only have gotten into that line if you had done a G to
        ///            append hold to pattern, or if you had done an x to swap hold and pattern
        ///            after having done an H).
        ///    d    -- print the current line and stop processing it.
        ///    D    -- stop processing this input line without printing it
        ///    q    -- print the current input line and process no more lines
        ///    Q    -- do not print the current input line and process no more lines
        ///    h    -- copy pattern to hold
        ///    H    -- append the pattern buffer to the hold buffer (\\n is the delimiter)
        ///    g    -- replace the pattern buffer with the hold buffer
        ///    G    -- append the hold buffer to the pattern buffer with \\n as a delimiter
        ///    x    -- swap the pattern and hold buffers
        ///
        ///@endcode
        ///
        ///@note The pattern buffer is the place where the current of input is stored at the
        ///      beginning of processing.  The hold buffer is a second string variable which
        ///      can be quickly swapped with the pattern buffer.  Two buffers are all you get.
        ///
        ///   
    {   // struct Statement

         std::list<Statement *>  childStatements_;
         ActivationMap           childStatementActivationMap_;

         Statement *parentStatement_;  // parent owns me, not the other way around

         Statement(Statement *parentStatement)  // null parent only ok for very top level statement
         : parentStatement_(parentStatement)
         {
         }

         LineRangeActivationState activationState() ;

         void setActivationState(LineRangeActivationState newState);

         virtual bool execute(ScriptRuntimeContext *context) = 0; 
            ///<  Execute this statement on the specified context.
            ///
            ///   @return true to indicate that the script should stop immediately.

         void clear()
         {
            CXXTLS_FOREACH(Statement *cur, childStatements_)
            {
               delete cur;
            }

            childStatements_.clear();
         }

         void adoptChild(Statement *newbie)
             ///
             ///  Take ownership of a new statement and schedule it for delete when *this is
             ///  destructed.
             ///
         {
             childStatements_.push_back(newbie);
         }


         virtual ~Statement()
         {
            clear();
         }

         std::string debugPrefix() const
         {
             char buffer[256];

             snprintf(buffer, sizeof(buffer), "ST[%p]=", this);

             return buffer;
         }


         virtual std::string debugString() const
         {
            return "<invalid>";
         }

         static void debugPrint(Statement *p)
         {
            printf("%s\n", p->debugString().c_str());
            fflush(stdout);
         }


    
    };  // struct statement
//========================================================================================
    struct OuterStatement
    :  public Statement
        /// A fairly useless kind of statement that exists only to serve as the
        /// top level statement in a script.  It is not "block" statement because
        /// it is not filterable.
    {
        OuterStatement()
        :  Statement(0)
        {
        }


        virtual bool execute(ScriptRuntimeContext *context);


         virtual std::string debugString() const
         {
            return "<Outer>";
         }




    };
//========================================================================================

    struct FilterableStatement
    :  public Statement
        /// A kind of sed Statement that can have 0, 1, or 2 filter criteria.
        ///
        /// With zero filter criteria, the statement is always executed.
        /// With one filter, the statement is only executed if the filter matches.
        /// With 2 filters, the statement is only executed on a range of lines
        /// that is defined by the first and second filters.  That is, if the
        /// first filter has passed on this or some line before it and the second
        /// filter has NOT passed on some earlier line.  That, is, the second
        /// filter defines the end of a group of lines.  The statement on that line
        /// will be executed, but none after it.
        ///
        /// Note that any statement that has filters, can have the logic of the
        /// filter range inverted.  Here are some examples of filters:
        ///
        ///@code
        ///
        ///   s/fred/bill/g     # unfiltered
        ///
        ///   10s/fred/bill/g   # on line 10, replace fred with bill, globally
        ///
        ///   $p                # only on the last line of the input, print the line
        ///
        ///   10,20p            # print only lines 10 through 20, inclusive
        ///
        ///   10!p              # print every line EXCEPT 10 (inverted logic case)
        ///
        ///   /^begin/,/^end/p  # print only lines in a begin/end block
        ///
        ///@endcode
        ///
        ///@note  The handling of filter result inversion is not part of the
        ///       filter class but part of the FilterableStatement class
    {  // struct FilterableStatement

       virtual ~FilterableStatement(){}

       int filters_;

       Filter filter1_;
       Filter filter2_;
       bool   inverted_;


       FilterableStatement(Statement*         parent,
                           int                fc, 
                           Filter const      &f1, 
                           Filter const      &f2,
                           bool               inverted
                          )
        :  Statement(parent)
        ,  filters_(fc)
        ,  filter1_(f1)
        ,  filter2_(f2)
        ,  inverted_(inverted)
        {
        }


        bool active(ScriptRuntimeContext *context, Statement *statement);  
            ///<  return true if the patternBuffer, line number, and
            ///   isLast_ flag in the specified context match the
            ///   requirements of the filters.
            ///
            ///   @param[in] context  The runtime context in which to
            ///                       determine if the current statement,
            ///                       as defined both by the statement pointer
            ///                       and by line number information in the
            ///                       context, is active.
            ///   @param[in] statement  The statement about to be executed on
            ///                         the current line of input as defined
            ///                         by the line number and is last flag
            ///                         in the context.
            ///
            ///   Note that the runtime context keeps track of the activation
            ///   status of statements, by the statement pointer, based on the
            ///   filters in a filterable statement.  Whenever you have a range
            ///   filter, the activate function manages the activation state.

        virtual std::string debugStringPrefix() const;

            
    }; // struct FilterableStatement

//========================================================================================

    struct BlockStatement
    : public FilterableStatement
         /// A group of statements that executes a single statement with a single
         /// pair of line context filters -- optionally with a "while condition"
         /// that allows the block to be run multiple times on the same input
         /// line while the condition is true.
         ///
         /// Typically defined like one of the following
         ///
         ///@code
         ///  [rangespec] [ W /regex/ ] "{ <statementList> "}"
         ///@endcode
         ///
         /// Where [rangespec] is 0, 1, or 2 Filters with an optional trailing ! to indicate
         /// range inversion.  The <statementlist> looks like this:
         ///
         ///@code
         ///
         ///  [statement { ; statemeent } ]
         ///
         ///@endcode
         ///
         ///
         /// In the "while" case, the statement list will be repeatedly executed
         /// while the regex matches the pattern buffer.
         ///
         /// In the non-while case, the statement list will be executed only once.
    {

        std::auto_ptr<SimpleRegex>  whileCondition_;

        bool execute(ScriptRuntimeContext *context);

        BlockStatement(Statement *parent,
                       int filterCount,
                       Filter &fc1,
                       Filter &fc2,
                       bool    filterInverted,
                       SimpleRegex *condition
                      )
        :  FilterableStatement(parent, filterCount, fc1, fc2, filterInverted)
        ,  whileCondition_(condition)
        {
        }
           
         virtual std::string debugString() const;

         // note that there is no parse() function here because the compileStream() method, below(),
         // parses the block statement for you.


    }; // BlockStatement

//========================================================================================

    struct SubstStatement
    :  public FilterableStatement
        ///  A regex replacment statement from sed.  Typically specified something like
        ///  this:
        ///
        ///@code
        ///
        ///   s/target/replacement/options
        ///
        ///@endcode
        ///
        ///  Where options includes at least one of 
        ///
        ///  @li  1  -- only replace the FIRST match of the target string in the pattern
        ///             buffer with replacement.
        ///
        ///  @li  g  -- replace all of the matches of the target with the replacement.
        ///
        ///  and is optionally followed by 
        ///
        ///  @li i -- forces the comparison to be case insensitive.
        ///
        ///
        /// Note that the target string is a basic regular expression, with all of it's complexities,
        /// and that the replacement string can contain special character sequences to allow
        /// the inclusion of @b parts of the substrings matching the target expression.  See the
        /// online SED language manuals for examples.
        /// 
    {

        virtual ~SubstStatement(){}

        int count_;  // number of substitutions

        SimpleRegex target_;
        SimpleRegex replacement_;

        SubstStatement(Statement         *parent,
                       int                fc, 
                       Filter const      &f1,
                       Filter const      &f2,
                       bool               filtersInverted,
                       std::string const &tgt,
                       std::string const &rep,
                       int                count,
                       bool               insensitive
                      )
        : FilterableStatement(parent, fc, f1, f2, filtersInverted)
        , count_(count)
        , target_(tgt, (insensitive? "i" : ""))
        , replacement_(rep)
        {
        }

        bool execute(ScriptRuntimeContext *context);
            ///<  Execute this statemement in a given runtime context.
            ///
            ///  @param[in,out] context  The runtime context in which to execute
            ///                          this statement.
            ///  @return true if the parsing was successful.
            
        template<class StringIterator1, class StringIterator2, class CompiledScript>
        static Statement* parse(Statement             *parent,
                                int                   filterCount, 
                                Filter const          &f1, 
                                Filter const          &f2, 
                                bool                   inverted,
                                StringIterator1       &firstChar,
                                StringIterator2 const &lastChar,
                                size_t                 line,
                                CompiledScript        &script
                               )
            ///  Parse a SubstStatement from an input stream.
            ///
            ///  @return A statement pointer, or null
            ///
            ///  @param[in]  parent       A pointer to the statement that will own this one.
            ///
            ///  @param[in]  filterCount  A count of the range filters for this statement.  0 means
            ///                           that this statement applies to all lines.  1 means that this
            ///                           statement only applies to specific line either by line number
            ///                           within the input stream, or by any line in the stream matching
            ///                           a regular expression, or by the very last line of the stream.
            ///                           When filterCount == 2, then the script statement only applies
            ///                           to a range of lines within the input stream.  The range is defined
            ///                           in a manner documented by class muSED::Filter.
            ///
            /// @param[in]     f1         Filter1.
            /// @param[in]     f2         Filter2;
            /// @param[in]     inverted   Flag indicating that the f1/f2 logic is inverted.
            /// @param[in,out] firstChar  A pointer to the first character to be parsed.
            /// @param[in]     lastChar   The end of the range of characters defining a script fragment.
        {
        
            // parse the sed 's' command
        
            ++firstChar;
        
            if(firstChar == lastChar)
            {
               script.fail("missing / in s command", line);
               return 0;
            }
        
            char c = *firstChar++;
        
            if(    (c <= ' ') 
               ||  (c >= 'a' && c <= 'z')
               ||  (c >= 'A' && c <= 'Z')
               ||  (c >= '0' && c <= '9')
              )
            {
               std::string tmp = "invalid s command delimiter '";
               tmp += c;
               tmp += "'";
        
               script.fail(tmp, line);
               return 0;
        
            }
        
            char delim = c;
        
            std::string target;
            std::string replacement;
            int count=0;
            bool insensitive=false;  // case insensitive search/replace
        
            // we are currently here in the parsing of an s command:
            //
            //   s/target/replacement/1
            //     ^
        
            if(!parseEscapedString(delim, target, firstChar, lastChar))
            {
               script.fail("badly formed LHS in s command", line);
               return 0;
            }


        
            ++firstChar;
        
            // we are currently here in the parsing of an s command:
            //
            //   s/target/replacement/1
            //            ^
        
            
            if(!parseEscapedString(delim, replacement, firstChar, lastChar))
            {
               script.fail("badly formed RHS in s command", line);
               return 0;
            }
        
            ++firstChar;

        
            if(firstChar == lastChar)
            {
               script.fail("badly formed count in s command -- expected 1 or g", line);
               return 0;
                
            }


            while((firstChar != lastChar) &&
                  (*firstChar == '1' ||
                   *firstChar == 'g' ||
                   *firstChar == 'i'
                  )
                 )
            {

                // 'g' is the default count character

                if(*firstChar == '1')
                   count=1;
                else
                if(*firstChar == 'i')
                {
                   insensitive = true;
                }

                ++firstChar;  // skip the g or the 1

            }

        
            expandEscapes(target);       // replace \n in target with newline, and so on.
            expandEscapes(replacement);  // replace \n in target with newline, and so on.


            return new SubstStatement(   parent,
                                         filterCount,
                                         f1,
                                         f2,
                                         inverted,
                                         target,
                                         replacement,
                                         count,
                                         insensitive
                                     );
        }

         virtual std::string debugString() const;


          
    }; // struct SubstStatement

//========================================================================================

    struct TranslateStatement
    :  public FilterableStatement
        ///  A character set replacment statement from sed.  Typically specified like
        ///  this:
        ///
        ///@code
        ///
        ///   y/target/replacement/
        ///
        ///@endode
        ///
        /// Note that this translate statement is more flexible than that found in
        /// SED or even gnu SED.  In the standard SED statement, the target and
        /// replacement must be simple character lists.  But in this statement,
        /// character translation ranges are allowed.  So, that this replacement
        ///@code
        ///
        ///   y/a-z/A-Z/
        ///
        ///@endcode
        ///
        /// Will actually translate all characters, a-z into their upper case form.
        /// This is the way the standard unix TR command works, but not the way the
        /// y command works in sed or even gnu SED.  In SED, the above y// command
        /// only translates a to A and z to Z (with - being translated into -).
        ///
        /// This significant difference is too valuable to ignore -- especially
        /// given the relative lack of utility of the SED command (and uncommon
        /// use since it is broken!)
    {

        virtual ~TranslateStatement(){}

        std::string target_;
        std::string replacement_;

        CharSetTranslator translator_;

        TranslateStatement(Statement         *parent,
                       int                fc, 
                       Filter const      &f1,
                       Filter const      &f2,
                       bool               filtersInverted,
                       std::string const &tgt,
                       std::string const &rep
                      )
        : FilterableStatement(parent, fc, f1, f2, filtersInverted)
        , target_(tgt)
        , replacement_(rep)
        , translator_(tgt,rep)
        {
        }

        bool execute(ScriptRuntimeContext *context);
            ///<  Execute this statemement in a given runtime context.
            ///
            ///  @param[in,out] context  The runtime context in which to execute
            ///                          this statement.
            ///  @param[in] map          Ignored.
            ///
            ///  @return true if the parsing was successful.
            
        template<class StringIterator1, class StringIterator2, class CompiledScript>
        static Statement* parse(Statement             *parent,
                                int                   filterCount, 
                                Filter const          &f1, 
                                Filter const          &f2, 
                                bool                   inverted,
                                StringIterator1       &firstChar,
                                StringIterator2 const &lastChar,
                                size_t                 line,
                                CompiledScript        &script
                               )
            ///  Parse a TranslateStatement from an input stream.
            ///
            ///  @return A statement pointer, or null
            ///
            ///  @param[in]  parent       A pointer to the statement that will own this one.
            ///
            ///  @param[in]  filterCount  A count of the range filters for this statement.  0 means
            ///                           that this statement applies to all lines.  1 means that this
            ///                           statement only applies to specific line either by line number
            ///                           within the input stream, or by any line in the stream matching
            ///                           a regular expression, or by the very last line of the stream.
            ///                           When filterCount == 2, then the script statement only applies
            ///                           to a range of lines within the input stream.  The range is defined
            ///                           in a manner documented by class muSED::Filter.
            ///
            /// @param[in]     f1         Filter1.
            /// @param[in]     f2         Filter2;
            /// @param[in]     inverted   Flag indicating that the f1/f2 logic is inverted.
            /// @param[in,out] firstChar  A pointer to the first character to be parsed.
            /// @param[in]     lastChar   The end of the range of characters defining a script fragment.
        {
        
            // parse the sed 'y' command
        
            ++firstChar;
        
            if(firstChar == lastChar)
            {
               script.fail("missing / in s command", line);
               return 0;
            }
        
            char c = *firstChar++;
        
            if(    (c <= ' ') 
               ||  (c >= 'a' && c <= 'z')
               ||  (c >= 'A' && c <= 'Z')
              )
            {
               std::string tmp = "invalid y command delimiter '";
               tmp += c;
               tmp += "'";
        
               script.fail(tmp, line);
               return 0;
        
            }
        
            char delim = c;
        
            std::string target;
            std::string replacement;
        
            // we are currently here in the parsing of an y command:
            //
            //   y/target/replacement/
            //     ^
        
            if(!parseEscapedString(delim, target, firstChar, lastChar))
            {
               script.fail("badly formed LHS in 'y' command", line);
               return 0;
            }
        
            ++firstChar;
        
            // we are currently here in the parsing of a 'y' command:
            //
            //   y/target/replacement/
            //            ^
        
            
            if(!parseEscapedString(delim, replacement, firstChar, lastChar))
            {
               script.fail("badly formed RHS in 'y' command", line);
               return 0;
            }
        
            ++firstChar;
        
        
            // hah!  good syntax
        
            expandEscapes(target);       // replace \n in target with newline, and so on.
            expandEscapes(replacement);  // replace \n in target with newline, and so on.


            return new TranslateStatement(   parent,
                                             filterCount,
                                             f1,
                                             f2,
                                             inverted,
                                             target,
                                             replacement
                                         );
        }

         virtual std::string debugString() const;


          
    }; // struct TranslateStatement

//========================================================================================

    struct SortStatement
    :  public FilterableStatement
        ///
        ///  Treat the pattern buffer as a collection of lines separated by newline
        ///  characters and re-order them in a sorted fashion based on the options
        ///  provided.
        ///
        ///@code
        ///
        ///   O/options/
        ///
        ///@endode
        ///
        ///  This command is best used afte you have accumulated lines into the
        ///  hold buffer, then swapped them back into the pattern buffer, in code
        ///  something like this:
        ///
        ///@code
        ///   H
        ///   ${
        ///      O/u/
        ///      p
        ///    }
        ///@endcode
        ///
        ///  This code fragment concatentes all the lines in the input stream into
        ///  the hold buffer, then after the last line has been accumulated, it
        //   sorts the lines and prints them.
    {

        virtual ~SortStatement(){}

        bool unique_;
        bool reverse_;
        bool badOptions_;

        SortStatement(Statement         *parent,
                       int                fc, 
                       Filter const      &f1,
                       Filter const      &f2,
                       bool               filtersInverted,
                       std::string const &options
                      )
        : FilterableStatement(parent, fc, f1, f2, filtersInverted)
        , unique_(false)
        , reverse_(false)
        , badOptions_(false)
        {
            CXXTLS_FOREACH(char c, options)
            {
                switch(c)
                {
                   case 'u':  unique_    = true; break;
                   case 'r':  reverse_   = true; break;
                   default:
                       badOptions_ = true;
                      
                }
            }

        }

        bool execute(ScriptRuntimeContext *context);
            ///<  Execute this statemement in a given runtime context.
            ///
            ///  @param[in,out] context  The runtime context in which to execute
            ///                          this statement.
            ///  @param[in] map          Ignored.
            ///
            ///  @return true if the parsing was successful.
            
        template<class StringIterator1, class StringIterator2, class CompiledScript>
        static Statement* parse(Statement             *parent,
                                int                   filterCount, 
                                Filter const          &f1, 
                                Filter const          &f2, 
                                bool                   inverted,
                                StringIterator1       &firstChar,
                                StringIterator2 const &lastChar,
                                size_t                 line,
                                CompiledScript        &script
                               )
            ///  Parse a SortStatement from an input stream.
            ///
            ///  @return A statement pointer, or null
            ///
            ///  @param[in]  parent       A pointer to the statement that will own this one.
            ///
            ///  @param[in]  filterCount  A count of the range filters for this statement.  0 means
            ///                           that this statement applies to all lines.  1 means that this
            ///                           statement only applies to specific line either by line number
            ///                           within the input stream, or by any line in the stream matching
            ///                           a regular expression, or by the very last line of the stream.
            ///                           When filterCount == 2, then the script statement only applies
            ///                           to a range of lines within the input stream.  The range is defined
            ///                           in a manner documented by class muSED::Filter.
            ///
            /// @param[in]     f1         Filter1.
            /// @param[in]     f2         Filter2;
            /// @param[in]     inverted   Flag indicating that the f1/f2 logic is inverted.
            /// @param[in,out] firstChar  A pointer to the first character to be parsed.
            /// @param[in]     lastChar   The end of the range of characters defining a script fragment.
        {
        
            // parse the sed 'O' command
        
            ++firstChar;   // skip 'O'
        
            if(firstChar == lastChar)
            {
               script.fail("missing / in O command", line);
               return 0;
            }
        
            char c = *firstChar++;
        
            if(    (c <= ' ') 
               ||  (c >= 'a' && c <= 'z')
               ||  (c >= 'A' && c <= 'Z')
              )
            {
               std::string tmp = "invalid O command delimiter '";
               tmp += c;
               tmp += "'";
        
               script.fail(tmp, line);
               return 0;
        
            }
        
            char delim = c;
        
            std::string target;
        
            // we are currently here in the parsing of an O command:
            //
            //   O/options/
            //     ^
        
            if(!parseEscapedString(delim, target, firstChar, lastChar))
            {
               script.fail("badly formed options in 'O' command", line);
               return 0;
            }
        
            ++firstChar;  // skip trailing /
        
            // we are currently here in the parsing of an 'O' command:
            //
            //   O/target/
            //            ^

            if(target.find_first_not_of("ur") < target.size())
            {
               script.fail("invalid option character in 'O' command", line);
               return 0;
            }

        
            return new SortStatement(   parent,
                                             filterCount,
                                             f1,
                                             f2,
                                             inverted,
                                             target
                                         );
        }

         virtual std::string debugString() const;


          
    }; // struct SortStatement

//========================================================================================

    struct PrintStatement
    :  public FilterableStatement
        ///
        /// A sed print statement, usually specified in one of the following forms:
        ///
        ///@code
        ///
        ///   p    # prints the whole pattern buffer
        ///   P    # prints only up to the first \\n.
        ///
        ///@endcode
        ///
        /// Prints the contents of the sed pattern buffer.  A "print" operation
        /// output one or more strings into the output sequence associated with
        /// the script's runtmie context.  See the muSED::ScriptRuntimeContext::print()
        ///
        /// Note that the two different forms of the statement behave
        /// slightly differently with respect to \\n characters in the pattern
        /// buffer.  In the unix sed executable, the pattern buffer will only have
        /// \\n characters in it if the script itself puts them there -- as lines are
        /// automatically split before script processing.  This input splitting does
        /// not occur in muSED.  The caller may include \\n in the input strings.
        /// 
        /// The script can force \\n characters into the pattern buffer in the following
        /// ways:
        ///
        ///   @li  the 'y' command can translate any character into a \\n
        ///   @li  the 's' command can replace any sequence of charactes into a \\n
        ///   @li  the 'G' command can append the hold buffer to the pattern buffer
        ///        using \\n as a delimiter
        ///   @li  the 'x' command can swap the pattern buffer with the hold buffer
        ///        which might have \\n in it if the H command were previously used
        ///        to append the pattern buffer to the hold buffer.
        ///
        /// The 'P' command prints only up to the first \\n in the pattern buffer to
        /// the output sequences (or the whole pattern buffer if there is no \\n).
        ///
        /// The 'p' command prints each \\n delimited component of the the pattern
        /// buffer to the output sequences.
    {

        virtual ~PrintStatement(){}

        bool firstLineOnly_;   ///< If true, print only up to the first \\n character
                               ///  found in the pattern buffer.   



        PrintStatement(Statement         *parent,
                       int                fc,
                       Filter const      &f1,
                       Filter const      &f2,
                       bool               filtersInverted,
                       bool               firstLineOnly
                      )
        : FilterableStatement(parent, fc, f1, f2, filtersInverted)
        , firstLineOnly_(firstLineOnly)
        {
        }


        bool execute(ScriptRuntimeContext *context);
            ///<  Invoke context->Print()

        template<class StringIterator1, class StringIterator2, class CompiledScript>
        static Statement* parse(Statement             *parent,
                                int                   filterCount, 
                                Filter const          &f1, 
                                Filter const          &f2, 
                                bool                   inverted,
                                StringIterator1       &firstChar,
                                StringIterator2 const &lastChar,
                                size_t                 line,
                                CompiledScript        &script
                               )
        {

            bool firstLineOnly = *firstChar == 'P';

            ++firstChar;  // skip the p

            return new PrintStatement(   parent,
                                         filterCount,
                                         f1,
                                         f2,
                                         inverted,
                                         firstLineOnly
                                     );
        }

         virtual std::string debugString() const;



          
    }; // struct PrintStatement

//========================================================================================

    struct SwapStatement
    :  public FilterableStatement
        ///
        /// A sed swap statement, usually specified like this:
        ///
        ///@code
        ///
        ///   x
        ///
        ///@endcode
        ///
        /// Swaps the pattern and hold buffer.
    {

        virtual ~SwapStatement(){}

        SwapStatement(Statement         *parent,
                       int                fc,
                       Filter const      &f1,
                       Filter const      &f2,
                       bool               filtersInverted
                      )
        : FilterableStatement(parent, fc, f1, f2, filtersInverted)
        {
        }


        bool execute(ScriptRuntimeContext *context);
            ///<  Invoke context->Print()

        template<class StringIterator1, class StringIterator2, class CompiledScript>
        static Statement* parse(Statement             *parent,
                                int                   filterCount, 
                                Filter const          &f1, 
                                Filter const          &f2, 
                                bool                   inverted,
                                StringIterator1       &firstChar,
                                StringIterator2 const &lastChar,
                                size_t                 line,
                                CompiledScript        &script
                               )
        {

            ++firstChar;  // skip the p

            return new SwapStatement(   parent,
                                         filterCount,
                                         f1,
                                         f2,
                                         inverted
                                     );
        }

         virtual std::string debugString() const;


          
    }; // struct SwapStatement


//========================================================================================

    struct DeleteStatement
    :  public FilterableStatement
        ///
        /// A sed delete statement, usually specified like this:
        ///
        ///@code
        ///
        ///   d
        ///
        ///@endcode
        ///
        ///
        /// Terminate the execution of the script with respect to the current line
        /// but not on the input stream as a whole.  The delete statement does not
        /// produce any output.  Subsequent input lines will be processed by the
        /// script as normal.
    {

        virtual ~DeleteStatement(){}

        DeleteStatement(Statement         *parent,
                       int                fc,
                       Filter const      &f1,
                       Filter const      &f2,
                       bool               filtersInverted
                      )
        : FilterableStatement(parent, fc, f1, f2, filtersInverted)
        {
        }


        bool execute(ScriptRuntimeContext *context);
            ///<  Invoke context->Print()

        template<class StringIterator1, class StringIterator2, class CompiledScript>
        static Statement* parse(Statement             *parent,
                                int                   filterCount, 
                                Filter const          &f1, 
                                Filter const          &f2, 
                                bool                   inverted,
                                StringIterator1       &firstChar,
                                StringIterator2 const &lastChar,
                                size_t                 line,
                                CompiledScript        &script
                               )
        {

            ++firstChar;  // skip the d

            return new DeleteStatement(   parent,
                                         filterCount,
                                         f1,
                                         f2,
                                         inverted
                                     );
        }

         virtual std::string debugString() const;


          
    }; // struct DeleteStatement


//========================================================================================

    struct QuitStatement
    :  public FilterableStatement
        ///
        /// A sed quit statement, usually specified like this:
        ///
        ///@code
        ///
        ///   q
        ///   Q
        ///
        ///@endcode
        ///
        ///
        /// Terminate the execution of the entire script on
        /// all of it's input.  The 'q' command, prints the current line and stops
        /// the script but the 'Q' command does not print the current line before 
        /// stopping.
    {
        bool printCurrent_;             ///< print current line before ending script


        virtual ~QuitStatement(){}

        QuitStatement(Statement         *parent,
                       int                fc,
                       Filter const      &f1,
                       Filter const      &f2,
                       bool               filtersInverted,
                       bool               printCurrent
                      )
        : FilterableStatement(parent, fc, f1, f2, filtersInverted)
        , printCurrent_(printCurrent)
        {
        }


        bool execute(ScriptRuntimeContext *context);
            ///<  Invoke context->Print()

        template<class StringIterator1, class StringIterator2, class CompiledScript>
        static Statement* parse(Statement             *parent,
                                int                   filterCount, 
                                Filter const          &f1, 
                                Filter const          &f2, 
                                bool                   inverted,
                                StringIterator1       &firstChar,
                                StringIterator2 const &lastChar,
                                size_t                 line,
                                CompiledScript        &script
                               )
        {
            bool printCurrent = (*firstChar) == 'q';

            ++firstChar;  // skip the q

            return new QuitStatement(   parent,
                                         filterCount,
                                         f1,
                                         f2,
                                         inverted,
                                         printCurrent
                                     );
        }

         virtual std::string debugString() const;


          
    }; // struct QuitStatement


//========================================================================================

    struct HoldStatement
    :  public FilterableStatement
        ///
        /// A sed hold statement, usually specified like this:
        ///
        ///@code
        ///
        ///   h             -- replace mode
        ///   H             -- append mode
        ///
        ///@endcode
        ///
        ///  Copy or append the current sed pattern buffer into the hold buffer.  A constructor
        ///  parameter, append, decides which.

    {

        bool append_;

        virtual ~HoldStatement(){}

        HoldStatement(Statement         *parent,
                       int                fc,
                       Filter const      &f1,
                       Filter const      &f2,
                       bool               filtersInverted,
                       bool               append=false
                      )
        : FilterableStatement(parent, fc, f1, f2, filtersInverted)
        , append_(append)
        {
        }


        bool execute(ScriptRuntimeContext *context);
            ///<  Invoke context->Print()

        template<class StringIterator1, class StringIterator2, class CompiledScript>
        static Statement* parse(Statement             *parent,
                                int                   filterCount, 
                                Filter const          &f1, 
                                Filter const          &f2, 
                                bool                   inverted,
                                StringIterator1       &firstChar,
                                StringIterator2 const &lastChar,
                                size_t                 line,
                                CompiledScript        &script,
                                bool                   append
                               )
        {

            ++firstChar;  // skip the p

            return new HoldStatement(   parent,
                                         filterCount,
                                         f1,
                                         f2,
                                         inverted,
                                         append
                                     );
        }


         virtual std::string debugString() const;


          
    }; // struct HoldStatement


//========================================================================================

    struct GetStatement
    :  public FilterableStatement
        ///
        /// A sed get statement, usually specified like this:
        ///
        ///@code
        ///
        ///   g             -- replace mode
        ///   G             -- append mode
        ///
        ///@endcode
        ///
        ///  Copy or append the current sed hold buffer into the pattern buffer.  A constructor
        ///  parameter, append, decides which.

    {

        bool append_;

        virtual ~GetStatement(){}

        GetStatement(Statement         *parent,
                       int                fc,
                       Filter const      &f1,
                       Filter const      &f2,
                       bool               filtersInverted,
                       bool               append=false
                      )
        : FilterableStatement(parent, fc, f1, f2, filtersInverted)
        , append_(append)
        {
        }


        bool execute(ScriptRuntimeContext *context);
            ///<  Invoke context->Print()

        template<class StringIterator1, class StringIterator2, class CompiledScript>
        static Statement* parse(Statement             *parent,
                                int                   filterCount, 
                                Filter const          &f1, 
                                Filter const          &f2, 
                                bool                   inverted,
                                StringIterator1       &firstChar,
                                StringIterator2 const &lastChar,
                                size_t                 line,
                                CompiledScript        &script,
                                bool                   append
                               )
        {

            ++firstChar;  // skip the p

            return new GetStatement(   parent,
                                         filterCount,
                                         f1,
                                         f2,
                                         inverted,
                                         append
                                     );
        }

         virtual std::string debugString() const;


          
    }; // struct GetStatement


//========================================================================================

    struct CompiledScript
        /// A list of compiled muSED statements and the parsing tools needed to
        /// create it.
        ///
        /// Such statements can be executed on a given line in a given
        /// ScriptRuntimeContext.
        ///
        /// The job of executing statements is broken up between the CompiledScript
        /// object and the various statements that it contains.  A statement handles
        /// the job of performing a single function on a single line, and the
        /// CompiledScript handles multiple lines, error reporting, etc.

    {

         OuterStatement  outerStatement_; // this statement alone has a null parent

         void clear()
         {
             outerStatement_.clear();
         }

         ~CompiledScript()
         {
            clear();
         }


         bool ok_;  // script has valid contents.

         std::string error_;


         template<class StringContainer>
         CompiledScript(StringContainer const &script)
         : ok_(false)
         , error_("Error:  no script specified")
         {

             compile(script.begin(), script.end());

         }

         bool ok() const { return ok_; }

         std::string const &error() const { return error_; }

         bool execute(ScriptRuntimeContext *context)
         {
             // execute all statements on the specified line

             if(!ok_)
                return true; // can't execute trash


             return outerStatement_.execute(context);
         }


         void fail(std::string const &errormsg, int line)
            /// Set the script parsing state to fail at a certain line
         {
            ok_    = false;

            char buffer[40];

            snprintf(buffer, sizeof(buffer), " at line %d",  line);


            error_ = errormsg + buffer;

            clear();  // it must all compile or none of it does.
            
         }

         typedef std::pair<std::string::const_iterator&, std::string::const_iterator&> CharRange;


#ifdef USE_muSED
         template<class LineRange>
         static void compileStream(  LineRange lineRange,                  // pair of REFERENCES to iterators
                                     CharRange charRange,                  // pair of REFERENCES to iterators
                                     Statement *parent,                    // owner of statements to be created
                                     int &line,
                                     CompiledScript *script
                                  )
         {

            ///  On entry, the charRange should refer to some segment of
            ///  the first line defined in the lineRange.  That is, charRange
            ///  must be valid pointers into lineRange.first.


             /// Parse statements from the specified range of
             /// strings in some container and produce
             /// statements into the statements_ member variable.

             std::string::const_iterator &firstChar = charRange.first;   // aliases to charRange references
             std::string::const_iterator &lastChar  = charRange.second;


             while(lineRange.first != lineRange.second)
             {
                // parse all lines in the script

                ++line;


                Filter filter1;
                Filter filter2;
                int    filterCount=0;

                while(firstChar != lastChar)
                {
                    // process all characters on the current line, with the
                    // possibility that if we hit a new line within an unclosed
                    // curly-brace block that the current line will be replaced
                    // with the next line, ad infinitum, until we hit the end
                    // of the curly block and all it's nested components...


                    skipBlanks(firstChar, lastChar);

                    if(firstChar == lastChar)
                       break;


                    if( *firstChar == ';' || *firstChar == '\n')
                    {
                       ++firstChar;
                       continue;
                    }


                    if(filterCount == 0)
                    {
                        if(filter1.parse(firstChar, lastChar))
                        {
                           ++filterCount;

                           if(firstChar != lastChar && *firstChar == ',')
                           {
                               ++firstChar;  // skip the comma

                               if(filter2.parse(firstChar, lastChar))
                                  ++filterCount;

                               if(filterCount != 2)
                               {
                                  script->fail("Error, expected filter expression after ','", line);
                               }
                           }
                        }
                    }



                    // filterCount will tell us if there are any filters
                    // at all.

                    // now parse the statement

                    skipBlanks(firstChar, lastChar);

                    bool filterInverted = false;

                    if(firstChar != lastChar && *firstChar == '!')
                    {
                        filterInverted = true;
                        ++firstChar;
                        skipBlanks(firstChar, lastChar);
                    }


                    if(firstChar != lastChar)
                    { // there is a command still on this line


                       // we have a statement, now parse it!

                       char cmd = *firstChar;

                       // NOTE:  the following code is stupid.
                       // Since all sed commands have a unique 1 character identifier,
                       // we should have a table matching that character with the
                       // kind of statement it matches, and let the parsing code
                       // reside in the statement itself.  As it is, this function is
                       // too big and is coupled to statement contstructors...


                       if(cmd == '#')
                       {
                          // this line is a comment
                          firstChar = lastChar;
                       }
                       else
                       if(cmd ==  's')
                       {

                         std::auto_ptr<Statement> s(SubstStatement::parse(parent,
                                                                          filterCount,
                                                                          filter1,
                                                                          filter2,
                                                                          filterInverted,
                                                                          firstChar,
                                                                          lastChar,
                                                                          line,
                                                                          *script
                                                    )
                                              );
                         if(s.get())
                         {
                             parent->adoptChild(s.release());
                         }  
                         else
                         {
                            return;
                         }

                       }
                       else
                       if(cmd ==  'y')
                       {

                         std::auto_ptr<Statement> s(TranslateStatement::parse(parent,
                                                                          filterCount,
                                                                          filter1,
                                                                          filter2,
                                                                          filterInverted,
                                                                          firstChar,
                                                                          lastChar,
                                                                          line,
                                                                          *script
                                                    )
                                              );
                         if(s.get())
                         {
                             parent->adoptChild(s.release());
                         }  
                         else
                         {
                            return;
                         }

                       }
                       else
                       if(cmd ==  'O')
                       {

                         std::auto_ptr<Statement> s(SortStatement::parse(parent,
                                                                         filterCount,
                                                                          filter1,
                                                                          filter2,
                                                                          filterInverted,
                                                                          firstChar,
                                                                          lastChar,
                                                                          line,
                                                                          *script
                                                                        )
                                                   );
                         if(s.get())
                         {
                             parent->adoptChild(s.release());
                         }  
                         else
                         {
                            return;
                         }

                       }
                       else
                       if( cmd == 'p' || cmd == 'P')
                       {
                                                 
                           std::auto_ptr<Statement> s(PrintStatement::parse(parent,
                                                                            filterCount,
                                                                            filter1,
                                                                            filter2,
                                                                            filterInverted,
                                                                            firstChar,
                                                                            lastChar,
                                                                            line,
                                                                            *script
                                                                            )
                                                     );
                           
                          if(s.get())
                          {
                              parent->adoptChild(s.release());
                          }
                          else
                          {
                              return;
                          }

                       }
                       else
                       if( cmd == 'd')
                       {
                                                 
                           std::auto_ptr<Statement> s(DeleteStatement::parse(parent,
                                                                            filterCount,
                                                                            filter1,
                                                                            filter2,
                                                                            filterInverted,
                                                                            firstChar,
                                                                            lastChar,
                                                                            line,
                                                                            *script
                                                                            )
                                                     );
                           
                          if(s.get())
                          {
                              parent->adoptChild(s.release());
                          }
                          else
                          {
                              return;
                          }

                       }
                       else
                       if( cmd == 'q' || cmd == 'Q' )
                       {
                                                 
                           std::auto_ptr<Statement> s(QuitStatement::parse(parent,
                                                                            filterCount,
                                                                            filter1,
                                                                            filter2,
                                                                            filterInverted,
                                                                            firstChar,
                                                                            lastChar,
                                                                            line,
                                                                            *script
                                                                            )
                                                     );
                           
                          if(s.get())
                          {
                              parent->adoptChild(s.release());
                          }
                          else
                          {
                              return;
                          }

                       }
                       else
                       if( cmd == 'h'  || cmd == 'H' )
                       {
                                                 
                           std::auto_ptr<Statement> s(HoldStatement::parse(parent,
                                                                            filterCount,
                                                                            filter1,
                                                                            filter2,
                                                                            filterInverted,
                                                                            firstChar,
                                                                            lastChar,
                                                                            line,
                                                                            *script,
                                                                            cmd == 'H'
                                                                            )
                                                     );
                           
                          if(s.get())
                          {
                              parent->adoptChild(s.release());
                          }
                          else
                          {
                              return;
                          }

                       }
                       else
                       if( cmd == 'g'  || cmd == 'G' )
                       {
                                                 
                           std::auto_ptr<Statement> s(GetStatement::parse(parent,
                                                                            filterCount,
                                                                            filter1,
                                                                            filter2,
                                                                            filterInverted,
                                                                            firstChar,
                                                                            lastChar,
                                                                            line,
                                                                            *script,
                                                                            cmd == 'G'
                                                                            )
                                                     );
                           
                          if(s.get())
                          {
                              parent->adoptChild(s.release());
                          }
                          else
                          {
                              return;
                          }

                       }
                       else
                       if( cmd == 'x')
                       {
                                                 
                           std::auto_ptr<Statement> s(SwapStatement::parse(parent,
                                                                            filterCount,
                                                                            filter1,
                                                                            filter2,
                                                                            filterInverted,
                                                                            firstChar,
                                                                            lastChar,
                                                                            line,
                                                                            *script
                                                                            )
                                                     );
                           
                          if(s.get())
                          {
                              parent->adoptChild(s.release());
                          }
                          else
                          {
                              return;
                          }

                       }
                       else
                       if(cmd == '{' || cmd == 'W')
                       {
                          // handle block statements here, recursively

                         std::auto_ptr<SimpleRegex> regex;

                         if(cmd == 'W')
                         {
                             // parse the regex

                             ++firstChar;  // skip the W

                             skipBlanks(firstChar, lastChar);

                             char c = ' ';

                             if(firstChar != lastChar)
                                 c = *firstChar;

                             if(    (c <= ' ') 
                                ||  (c >= 'a' && c <= 'z')
                                ||  (c >= 'A' && c <= 'Z')
                               )
                             {
                                std::string tmp = "invalid W command regex delimiter '";
                                tmp += c;
                                tmp += "'";
                         
                                script->fail(tmp, line);
                                return ;
                         
                             }

                             std::string target;
                             
                             // we are currently here in the parsing of an s command:
                             //
                             //     /regex/
                             //     ^
                             
                             ++firstChar;
                             if(!parseEscapedString(c, target, firstChar, lastChar))
                             {
                                script->fail("badly formed W command regex", line);
                                return;
                             }

                             ++firstChar;

                             // we are currently here in the parsing of an s command:
                             //
                             //     /regex/
                             //            ^

                             skipBlanks(firstChar, lastChar);

                             if(firstChar == lastChar || *firstChar != '{')
                             {
                                script->fail("Missing '{' after 'W' command regex", line);
                                return ;
                             }

                             expandEscapes(target);
                             regex.reset( new SimpleRegex(target) );

                         }


                         ++firstChar;

                         Statement *childBlock;

                         parent->adoptChild( childBlock =
                                             new BlockStatement(parent,
                                                              filterCount,
                                                              filter1,
                                                              filter2,
                                                              filterInverted,
                                                              regex.release()
                                                             )
                                                );

                         // note:  lineRange and charRange are pairs of references to
                         // iterators, so the call below passes this function's input
                         // references to its child.  

                         compileStream(lineRange, charRange, childBlock, line, script);

                         if( (firstChar == lastChar) || *firstChar != '}')
                         {
                            if(script->ok())
                                script->fail("block statement missing }", line);

                            return;
                         }

                         ++firstChar;



                       }
                       else
                       if(cmd == '}')
                       {
                          // although this might be an error, treat the }
                          // character as an end of script flag.  This is
                          // needed or at least helpful to recursive block
                          // parsing.

                          return;

                       }
                       else
                       {
                           std::string tmp = "unexpected command: ";

                           tmp += cmd;

                           script->fail(tmp, line);

                           return;
                          
                       }



                    } // there is a command still on this line


                    // by this point, the statement on the currentline has been processed,
                    // but we must deal with whatever comes after the valid syntax on the line.
                    // we will handle it by skipping blanks and then interpreting what we 
                    // see.

                    skipBlanks(firstChar, lastChar);

                    if(firstChar != lastChar)
                    { // there may be more on this same line

                        if(*firstChar == ';')
                        {
                           ++firstChar;     // the skip the ';'
                           filterCount=0;

                        }
                        else
                        if(*firstChar == '#')
                        {
                           firstChar = lastChar;  // end of line comment
                        }

                       
                    } // there may be more on this same line


                    
                }  // process the current line






                ++lineRange.first;  // step to next line

                if(lineRange.first != lineRange.second)
                {
                   firstChar = lineRange.first->begin();
                   lastChar  = lineRange.first->end();
                }


             }
             
         }
#endif


         template<class StringIterator>
         void compile(StringIterator first, StringIterator last)
         {
             if(first == last)
               return;  // leave the CompiledScript in the current
                        // error state

             ok_    = true;
             error_ = "";

             int line=0;

             std::string::const_iterator firstChar = first->begin(),
                                         lastChar  = first->end();

             typedef std::pair<StringIterator&, StringIterator&>                           LineRange;

             compileStream( LineRange(first,last), 
                            CharRange(firstChar,lastChar),
                            &outerStatement_, 
                            line, 
                            this 
                          );

         }


    };  // struct CompiledScript

//========================================================================================

    struct ScriptRuntimeContext
       ///  The runtime data needed to execute a muSED script on a stream of
       ///  input lines.  It contains the script and the activation state of the
       ///  lines within the script, as well as the pattern and hold buffers, the
       ///  only two variables defined by the sed language.
       ///
       ///  The ScriptRuntimeContext also defines functions needed to manipulate
       ///  the program variables so as to produce useful output.
       ///
       ///  The SED script language makes heavy use of the pattern buffer, which refers 
       ///  to the most recently read line of text from the input stream being processed
       ///  by the script.  The hold buffer is a temporary string available to the script
       ///  for keeping track of data previously read.  The hold buffer can be as simple
       ///  as an earlier line or it can be sum of all previous lines, or it can also contain
       ///  a collection of "variables" which refer to part of the previous lines read.
       ///  The user has the option of formatting the "hold" buffer in any helpful way
       ///  to a specific script.
       ///
       ///  The while command is somewhat dangerous in that it allows multiple transformations
       ///  on the same input command line.  It is dangerous because it allows you create
       ///  an infinite loop.  To avoid this, there is a maximum while loop iteration
       ///  count.  This can be set using the setMaxWhileIterations() function.  The
       ///  default is 1 million.  This is a large number but won't allow a script to run
       ///  forever.  Remember that this number applies to each line of input.  If you 
       ///  really need more than a million loops you are probably doing something wrong.
       ///  Text editors using muSED will likely set a much lower number.
    {
       CompiledScript &script_;

       bool running_;                               ///< The script is running and not yet been 'q' commanded.


       std::list<std::string> *output_;             ///< null unless we are in the
                                                    ///  execute function.  NOT OWNED
                                                    ///  BY THIS CLASS
       size_t lineNumber_;                          ///< Current line number within the input stream
                                                                                                    
       bool   isLast_;                              ///< True only if the current line number is the last
                                                    ///  in the input stream


       std::map<Statement*, bool>  activationMap_;  ///< A mapping between statements
                                                    ///  and their activation state.
                                                    ///  A statement is active unless it's
                                                    ///  filters turn it off.  Only
                                                    ///  2 filter statements can ever be
                                                    ///  off.  One filter statements are
                                                    ///  on until they are off, and then
                                                    ///  stay off only for the execution of
                                                    ///  one line.

       std::string buffer1_, buffer2_;              ///< "pattern" and "hold" in sed lingo

       std::string *pattern_;                       ///< defines which buffer is pattern
                                                    ///  and which is hold

       int maxWhileIterations_;                     ///< Maximum number of times a while
                                                    ///  command should loop on the same
                                                    ///  line.  

       ScriptRuntimeContext(CompiledScript &s)
       : script_(s)
       , running_(false)
       , output_(0)
       , lineNumber_(0)
       , isLast_(false)
       , pattern_(&buffer1_)
       , maxWhileIterations_(1000 * 1000)
       {
       }

       void stop() 
       {
          running_ = false;  // the apply() functions set this to true
       }

       void start() 
       {
          running_ = true;
       }

       bool running() const
       {
           return running_;
       }


       bool ok() const { return script_.ok(); }     ///< return true if this script's context is 
                                                    ///  valid.

       std::string const &error() const { return script_.error(); }
           ///<  Return a non-empty string if there has been a compile error in this
           ///   script.

       std::string *patternBuffer()
           ///  Return a pointer to this runtime context's current pattern buffer.
           ///
           ///   @note that the pattern and the hold buffer switch back and forth based on the
           ///   number of the 'x'  commands that have been executed in this script.
       {
           return pattern_;
       }

       std::string *holdBuffer()
           /// returns a pointer to the buffer that is not the pattern buffer.
       {


           if(pattern_ == &buffer1_)
              return &buffer2_;

           return &buffer1_;
       }

       void bufferSwap()
           /// swap pattern and hold
       {
          /// swap pattern and hold buffers

          if(pattern_ == &buffer1_)
             pattern_ = &buffer2_;
          else
             pattern_ = &buffer1_;
         
       }

       void hold()
           /// copy pattern to hold
       {
           *holdBuffer() = *patternBuffer();
       }

       void Hold()
           /// append pattern to hold
       {
          // a script will typically not execute H until it has first executed h

          *holdBuffer() += '\n';
          *holdBuffer() += *patternBuffer();
       }

       void get()
           /// copy hold to pattern
       {
          *patternBuffer() = *holdBuffer();
       }

       void Get()
           /// append hold to pattern
       {
          *patternBuffer() += '\n';
          *patternBuffer() += *holdBuffer();
       }

       int maxWhileIterations() const { return maxWhileIterations_; }
           ///<  @return the maximum number of times a while statement
           ///           is allowed to process a single line.  The
           ///           default is 1 million. 

       void setMaxWhileIterations(int count) { maxWhileIterations_ = count; }
           ///<  Set the maximum number of times that a while loop is allowed
           ///   to iterate over the same line of input text.  The default is
           ///   1 million but that is probably too large if you are 
           ///   applying a script from inside a text editor based on user
           ///   command.




       void print(bool firstLineOnly);  ///< Print part or all of the pattern buffer to the output.
                                        ///  Each newline delimited fragment goes into a separate
                                        ///  string in the output sequence.
                                        ///  @param[in] firstLineOnly  Means to print only the
                                        ///                            first newline delimited
                                        ///                            string to the output.

       bool execute(std::string const &currentLine, 
                    size_t             lineNumber, 
                    bool               isLast, 
                    std::list<std::string> &tmp,
                    ActivationMap          *activationMap
                   )
       {
          if(!ok())
              return false;

          output_          = &tmp;

          *patternBuffer() = currentLine;
          lineNumber_      = lineNumber;
          isLast_          = isLast;

          running_ = true;  // the q command will call stop() to set this to false.

          return script_.execute(this);
       }


    };  // struct ScriptRuntimeContext

//========================================================================================

    template<class StringContainer1, class StringContainer2>
    std::string apply(CompiledScript          &script,
                      StringContainer1 const  &input,
                      StringContainer2        &output,
                      int                      maxIter=0
                     )
    {
      /// apply a script to the strings in the input container
      /// and store the results in the output container.
      ///
      /// if the script is "not ok" then don't perform any
      /// operations and return  false;
      ///
      ///  @return std::string an error message if script compile
      ///                      error occurred.
      ///  
      ///  @param[in]  script   A string container holding the script.
      ///  @param[in]  input    The input data stream to process.
      ///  @param[out] output   The output data set.
      ///  @param[in]  maxIter  The maximum number of times a while loop
      ///                       will be allowed to iterate on a single
      ///                       line of input.  0 signifies that the
      ///                       default value should be used.  The default
      ///                       can be found in muSED::ScriptRuntimeContext.



      ScriptRuntimeContext context(script);
      ActivationMap        activationMap;

      if(!context.ok())
          return context.error(); //  check the script's error if you get false here
      

      if(maxIter)
         context.setMaxWhileIterations(maxIter);

      std::string currentLine;

      size_t lineNo = 1;



      output.clear();


      typename StringContainer1::const_iterator first =input.begin(),
                                                last  = input.end();

      if(first == last)
         return std::string();  // success

      context.start();


      do
      {
          currentLine =  *first++;

          bool eof = (first == last);

          if(apply(context, currentLine, lineNo, eof, output, &activationMap))
          {
             if(!context.running())
                  break;
          }

          ++lineNo;

      }
      while(first != last);

      return std::string();  // success

    }  // apply()

//========================================================================================

    template<class ScriptContainer, class StringContainer1, class StringContainer2>
    std::string apply(ScriptContainer const  &scriptText, 
                      StringContainer1 const &input,
                      StringContainer2       &output,
                      int                     maxIter=0
                     )
        /// Easy interface for a multi-line script -- compile it and execute it
        /// on the input and output containers.  This interface will be slower than
        /// compiling the script once and using the compiled script many times, but
        /// has a simpler interface.
        ///
        ///  @param[in]  scriptText  The script as a sequences of lines of text.
        ///  @param[in]  input       The text to be transformed by the script.
        ///  @param[ou]  output      The sequence of output text.
        ///  @param[in]  maxIter     The maximum number of times a while loop
        ///                          will be allowed to iterate on a single
        ///                          line of input.  0 signifies that the
        ///                          default value should be used.  The default
        ///                          can be found in muSED::ScriptRuntimeContext.

    {
        CompiledScript script(scriptText);

        if(!script.ok())
          return script.error();

        apply(script, input, output, maxIter);

        return std::string();

    }

//========================================================================================

    template<class StringContainer1, class StringContainer2>
    std::string oneLiner(std::string      const &scriptText, 
                         StringContainer1 const &input,
                         StringContainer2       &output,
                         int                     maxIter=0
                        )
        /// Easy interface for a single line script -- compile it and execute it
        /// on the input and output containers.  This interface will be slower than
        /// compiling the script once and using the compiled script many times, but
        /// has a simpler interface.
        ///
        /// Note that since ; is a command separator, as well as end of line, you
        /// can still have your script multiple commands.
        ///
        /// Sorry about the name. The template instantiator could not properly
        /// disambiguate all these "apply" methods without the name change.

    {
        std::list<std::string>  scriptContainer;

        scriptContainer.push_back(scriptText);

        CompiledScript script(scriptContainer);

        if(!script.ok())
            return script.error();

        apply(scriptContainer, input, output, maxIter);

        return std::string();

    }

//========================================================================================

   template<class StringContainer>
   bool apply(ScriptRuntimeContext &context, 
              std::string const    &currentLine,
              size_t                lineNumber,
              bool                  isLast,
              StringContainer      &output,
              ActivationMap        *activationMap
             )
       ///  Apply a line, with accompanying attributes, to the specified
       ///  runtime context, which includes a compiled script.  Basically this
       ///  is the function that executes a compiled script on a given line
       ///  to produce output specific to that line.  The assumption is that
       ///  if the script says not to return any output data, via the 
       ///  return code from the ScriptRuntimeContext::execute(), then it won't
       ///  have produced any output by mistake.
       ///
       ///  A given script invocation on a line can produce multiple lines of
       ///  output, if need be.
       ///
       ///  @param[in]  context   The runtime context for the script, including
       ///                        it's compiled form.
       ///
       ///  @param[in]  currentLine The text of the current line to be
       ///                          processed by the script.
       ///
       ///  @param[in]  isLast      A flag indicating that this particular
       ///                          line is the last one in the input stream.
       ///
       ///  @param[in]  output      A std::list&lgt;std::string&gt; that defines
       ///                          the script's output for this one line.
       ///  
       ///  @param[in,out] activationMap  The data structure that keeps track
       ///                                of whether the particular scritp
       ///                                statements are currently active on
       ///                                any given line.
   {
       
       std::list<std::string> tmp;

       bool rv = context.execute(currentLine, lineNumber, isLast, tmp, activationMap);

       CXXTLS_FOREACH(std::string const &outputString, tmp)
       {
          output.push_back(outputString);
       }

       return rv;

   }









}; // namespace muSED

}; // namespace cxxtls
